#ifndef _SHAPEOBJECTS_H
#define _SHAPEOBJECTS_H


#pragma pack(push, enter_shapefile_h)
#pragma pack(1)

inline size_t ByteSwitchInt(size_t n)
{
	return	((n & 0x000000FF) << 24) |
			((n & 0x0000FF00) << 8) |
			((n & 0x00FF0000) >> 8) |
			((n & 0xFF000000) >> 24);
}


// The different types of Shapefile objects.
// One object corresponds to exactly one entry in the dBase database
// table.
enum ShapeType
{
	Shape_NullShape		=  0,	// No shape at all.
	Shape_Point			=  1,	// Point, e.g. a village.
	Shape_Arc			=  3, 	// Arc, a collection of polylines.
	Shape_Polygon		=  5, 	// Polygon
	Shape_MultiPoint	=  8,	// A collection of points.
	Shape_PointZ		= 11,	// Point with Z and M.
	Shape_PolyLineZ		= 13,	// Polyline with Z and M
};


// Record header.
class CRecordHeader
{
public:
	static int Size() { return 8; }

	size_t GetRecordNumber() const { return ByteSwitchInt(m_recordNumber); }
	size_t GetContentLength() const { return ByteSwitchInt(m_contentLength); }

private:
	// The record header data.
	size_t m_recordNumber;
	size_t m_contentLength;
};


// Base class for all shapes
class CShapeRecord
{
public:
	struct XYPair
	{
		double x;
		double y;
	};

	// Returns the full size of this class, except any extra data appended at the end
	size_t      RecordSize()   const;

	// Returns size of data, not counting the record header nor the ShapeType
	size_t      DataSize()     const;

	// Get shape type
	ShapeType   GetShapeType() const { return m_shapeType.type; }

	// Get index of this shape
	size_t      GetIndex()     const { return m_header.GetRecordNumber() - 1; }

	// Bounding box data
	double GetXMin()      const;
	double GetYMin()      const;
	double GetXMax()      const;
	double GetYMax()      const;

	size_t        GetPartCount()   const;
	size_t        GetPointCount()  const;
	const XYPair& GetPoint(size_t idx) const;
	size_t        GetPart(size_t idx) const;

	double GetZ(size_t idx) const;
	double GetM(size_t idx) const;

	void* GetExtraDataPtr();

protected:
	CShapeRecord(); // Constructor not implemented. Should not be used.
	CRecordHeader m_header;
	union { // We use a union here to guarantee 4 bytes
		int IntShapeType;
		ShapeType type;
	} m_shapeType;
};



// The data for a Shapefile point object.
class CShapePointRecord : public CShapeRecord
{
protected:
	CShapePointRecord() {}
	~CShapePointRecord(); // No implementation. Don't use!
public:

	static CShapePointRecord* CreateRecord(void* pData);

	// Returns the full size of this class, except any extra data appended at the end
	int RecordSize() const { return sizeof(CShapeRecord) + DataSize(); }

	// Returns size of data, not counting the record header nor the ShapeType
	int DataSize() const { return 2*sizeof(double) +
							(m_shapeType.type == Shape_PointZ ? 2*sizeof(double) : 0); }

	// Shapefile point data.
	double GetX() const { return m_x; }
	double GetY() const { return m_y; }
	const XYPair* GetPoint() const { return reinterpret_cast<const XYPair*>(&m_x); }
	double GetZ() const { return (m_shapeType.type == Shape_PointZ ? *((&m_y)+1) : 0.0); }
	double GetM() const { return (m_shapeType.type == Shape_PointZ ? *((&m_y)+2) : 0.0); }

	// Bounding box data
	double GetXMin()      const { return m_x; }
	double GetYMin()      const { return m_y; }
	double GetXMax()      const { return m_x; }
	double GetYMax()      const { return m_y; }

	int    GetPartCount()   const { return 1; }
	int    GetPointCount()  const { return 1; }

	// The extra data pointer is only valid if the loader returned
	// an 'extra data size' > 0.
	void* GetExtraDataPtr() { return reinterpret_cast<char*>(this) + sizeof(CShapePointRecord); }

private:
	double m_x;
	double m_y;
};



// The data for a Shapefile arc object.
class CShapeArcRecord : public CShapeRecord
{
protected:
	CShapeArcRecord() {}
	~CShapeArcRecord(); // No implementation. Don't use!

public:
	static CShapeArcRecord* CreateRecord(void* pData);

	int RecordSize() const { return sizeof(CShapeRecord) + DataSize(); }

	int DataSize() const { return	4*sizeof(double) +
									2*sizeof(int) +
									GetPartCount() * sizeof(int) +
									GetPointCount() * sizeof(XYPair) +
									(m_shapeType.type == Shape_PolyLineZ ? (2+GetPointCount())*2*sizeof(double) : 0); }

	// Bounding box data
	double GetXMin()      const { return m_xMin; }
	double GetYMin()      const { return m_yMin; }
	double GetXMax()      const { return m_xMax; }
	double GetYMax()      const { return m_yMax; }

	double        GetZMin()     const { return m_shapeType.type != Shape_PolyLineZ ? 0. : *reinterpret_cast<const double*>(GetPoints() + GetPointCount()); }
	double        GetMMin()     const { return m_shapeType.type != Shape_PolyLineZ ? 0. : *(reinterpret_cast<const double*>(GetPoints() + GetPointCount()) + GetPointCount() + 2); }
	double        GetZMax()     const { return m_shapeType.type != Shape_PolyLineZ ? 0. : *(reinterpret_cast<const double*>(GetPoints() + GetPointCount()) + 1); }
	double        GetMMax()     const { return m_shapeType.type != Shape_PolyLineZ ? 0. : *(reinterpret_cast<const double*>(GetPoints() + GetPointCount()) + GetPointCount() + 3); }

	const double* GetZMinPtr()  const { return m_shapeType.type != Shape_PolyLineZ ? NULL : reinterpret_cast<const double*>(GetPoints() + GetPointCount()); }
	const double* GetZMaxPtr()  const { return m_shapeType.type != Shape_PolyLineZ ? NULL : reinterpret_cast<const double*>(GetPoints() + GetPointCount()) + 1; }

	const double* GetMMinPtr()  const { return m_shapeType.type != Shape_PolyLineZ ? NULL : reinterpret_cast<const double*>(GetPoints() + GetPointCount()) + GetPointCount() + 2; }
	const double* GetMMaxPtr()  const { return m_shapeType.type != Shape_PolyLineZ ? NULL : reinterpret_cast<const double*>(GetPoints() + GetPointCount()) + GetPointCount() + 3; }

	const XYPair& GetXMinYMin() const { return reinterpret_cast<const XYPair&>(*&m_xMin); }
	const XYPair& GetXMaxYMax() const { return reinterpret_cast<const XYPair&>(*&m_xMax); }

	size_t        GetPartCount()   const { return m_nPartCount; }
	size_t        GetPointCount()  const { return m_nPointCount; }
	const size_t* GetParts()       const { return &m_parts; }
	size_t        GetPart(size_t idx) const { return GetParts()[idx]; }
	const XYPair* GetPoints()   const { return reinterpret_cast<const XYPair*>(&m_parts + GetPartCount()); }
	const XYPair& GetPoint(size_t idx)   const { return GetPoints()[idx]; }

	const double* GetZs()   const { return m_shapeType.type == Shape_PolyLineZ ? reinterpret_cast<const double*>(GetPoints() + (1+GetPointCount())) : NULL; }
	const double& GetZ(size_t idx)   const { return GetZs()[idx]; }

	const double* GetMs()   const { return m_shapeType.type == Shape_PolyLineZ ? reinterpret_cast<const double*>(GetPoints() + (1+GetPointCount())) + GetPointCount() + 2 : NULL; }
	const double& GetM(size_t idx)   const { return GetMs()[idx]; }

	// The extra data pointer is only valid if the loader returned
	// an 'extra data size' > 0.
	void* GetExtraDataPtr() { return const_cast<XYPair*>(&GetPoint(GetPointCount())); }

private:
	// Shapefile arc data.
	double			   m_xMin;
	double			   m_yMin;
	double			   m_xMax;
	double			   m_yMax;
	size_t 			   m_nPartCount;
	size_t 			   m_nPointCount;
	size_t			   m_parts;
};


// The data for a Shapefile polygon object.
class CShapePolygonRecord : public CShapeArcRecord
{
protected:
	CShapePolygonRecord() {}
	~CShapePolygonRecord(); // No implementation. Don't use!
public:
};


#pragma pack(pop, enter_shapefile_h)

#endif // _SHAPEFILE_H
